# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""CPU implementations of ``n``-dimensional Cartesian spaces."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
from future.utils import native
standard_library.install_aliases()
from builtins import super

import ctypes
from functools import partial
from numbers import Integral
import numpy as np
import scipy.linalg as linalg
from scipy.sparse.base import isspmatrix

from odl.operator import Operator
from odl.set import RealNumbers, ComplexNumbers
from odl.space.base_ntuples import (
    NtuplesBase, NtuplesBaseVector, FnBase, FnBaseVector)
from odl.space.weighting import (
    WeightingBase, MatrixWeightingBase, VectorWeightingBase,
    ConstWeightingBase, NoWeightingBase,
    CustomInnerProductBase, CustomNormBase, CustomDistBase)
from odl.util.ufuncs import NumpyNtuplesUFuncs
from odl.util.utility import dtype_repr, is_real_dtype


__all__ = ('NumpyNtuples', 'NumpyNtuplesVector', 'NumpyFn', 'NumpyFnVector',
           'MatVecOperator',
           'NumpyFnMatrixWeighting', 'NumpyFnVectorWeighting',
           'NumpyFnConstWeighting',
           'npy_weighted_dist', 'npy_weighted_norm', 'npy_weighted_inner')


_BLAS_DTYPES = (np.dtype('float32'), np.dtype('float64'),
                np.dtype('complex64'), np.dtype('complex128'))


class NumpyNtuples(NtuplesBase):

    """Set of n-tuples of arbitrary type."""

    def element(self, inp=None, data_ptr=None):
        """Create a new element.

        Parameters
        ----------
        inp : `array-like`, optional
            Input to initialize the new element.

            If ``inp`` is ``None``, an empty element is created with no
            guarantee of its state (memory allocation only).

            If ``inp`` is a `numpy.ndarray` of shape ``(size,)``
            and the same data type as this space, the array is wrapped,
            not copied.
            Other `array-like` objects are copied.

        Returns
        -------
        element : `NumpyNtuplesVector`
            The new element created (from ``inp``).

        Notes
        -----
        This method preserves "array views" of correct size and type,
        see the examples below.

        Examples
        --------
        >>> strings3 = NumpyNtuples(3, dtype='U1')  # 1-char strings
        >>> x = strings3.element(['w', 'b', 'w'])
        >>> print(x)
        [w, b, w]
        >>> x.space
        ntuples(3, '<U1')

        Construction from data pointer:

        >>> int3 = NumpyNtuples(3, dtype='int')
        >>> x = int3.element([1, 2, 3])
        >>> y = int3.element(data_ptr=x.data_ptr)
        >>> print(y)
        [1, 2, 3]
        >>> y[0] = 5
        >>> print(x)
        [5, 2, 3]
        """
        if inp is None:
            if data_ptr is None:
                arr = np.empty(self.size, dtype=self.dtype)
                return self.element_type(self, arr)
            else:
                ctype_array_def = ctypes.c_byte * (self.size *
                                                   self.dtype.itemsize)
                as_ctype_array = ctype_array_def.from_address(data_ptr)
                as_numpy_array = np.ctypeslib.as_array(as_ctype_array)
                arr = as_numpy_array.view(dtype=self.dtype)
                return self.element_type(self, arr)
        else:
            if data_ptr is None:
                if inp in self:
                    return inp
                else:
                    arr = np.array(inp, copy=False, dtype=self.dtype, ndmin=1)
                    if arr.shape != (self.size,):
                        raise ValueError('expected input shape {}, got {}'
                                         ''.format((self.size,), arr.shape))

                    return self.element_type(self, arr)
            else:
                raise ValueError('cannot provide both `inp` and `data_ptr`')

    def zero(self):
        """Create a vector of zeros.

        Examples
        --------
        >>> c3 = NumpyNtuples(3, dtype=complex)
        >>> x = c3.zero()
        >>> x
        ntuples(3, 'complex').element([0j, 0j, 0j])
        """
        return self.element(np.zeros(self.size, dtype=self.dtype))

    def one(self):
        """Create a vector of ones.

        Examples
        --------
        >>> c3 = NumpyNtuples(3, dtype=complex)
        >>> x = c3.one()
        >>> x
        ntuples(3, 'complex').element([(1+0j), (1+0j), (1+0j)])
        """
        return self.element(np.ones(self.size, dtype=self.dtype))

    def __repr__(self):
        """Return ``repr(self)``."""
        constructor_name = 'ntuples'

        inner_str = '{}'.format(self.size)
        inner_str += ', {}'.format(dtype_repr(self.dtype))

        return '{}({})'.format(constructor_name, inner_str)

    @property
    def element_type(self):
        """`NumpyNtuplesVector`"""
        return NumpyNtuplesVector

    impl = 'numpy'

    @staticmethod
    def available_dtypes():
        """Available data types.

        Notes
        -----
        This is all dtypes available to numpy. See ``numpy.sctypes``
        for more information.

        The available dtypes may depend on the specific system used.
        """
        all_types = []
        for val in np.sctypes.values():
            all_types.extend(val)
        return all_types


class NumpyNtuplesVector(NtuplesBaseVector):

    """Representation of an `NumpyNtuples` element."""

    def __init__(self, space, data):
        """Initialize a new instance."""
        if not isinstance(space, NumpyNtuples):
            raise TypeError('{!r} not an `NumpyNtuples` instance'
                            ''.format(space))

        if not isinstance(data, np.ndarray):
            raise TypeError('`data` {!r} not a `numpy.ndarray` instance'
                            ''.format(data))

        if data.dtype != space.dtype:
            raise TypeError('`data` {!r} not of dtype {!r}'
                            ''.format(data, space.dtype))
        self.__data = data

        NtuplesBaseVector.__init__(self, space)

    @property
    def data(self):
        """Raw Numpy array representing the data."""
        return self.__data

    def asarray(self, start=None, stop=None, step=None, out=None):
        """Extract the data of this array as a numpy array.

        Parameters
        ----------
        start : int, optional
            Start position. ``None`` means the first element.
        stop : int, optional
            One element past the last element to be extracted.
            ``None`` means the last element.
        step : int, optional
            Step length. ``None`` is equivalent to 1.
        out : `numpy.ndarray`, optional
            Array to which the result should be written.
            Has to be contiguous and of the correct data type.

        Returns
        -------
        asarray : `numpy.ndarray`
            Numpy array of the same type as the space.

        Examples
        --------
        >>> import ctypes
        >>> vec = NumpyNtuples(3, 'float').element([1, 2, 3])
        >>> vec.asarray()
        array([ 1.,  2.,  3.])
        >>> vec.asarray(start=1, stop=3)
        array([ 2.,  3.])

        Using the out parameter

        >>> out = np.empty((3,), dtype='float')
        >>> result = vec.asarray(out=out)
        >>> out
        array([ 1.,  2.,  3.])
        >>> result is out
        True
        """
        if out is None:
            return self.data[start:stop:step]
        else:
            out[:] = self.data[start:stop:step]
            return out

    @property
    def data_ptr(self):
        """A raw pointer to the data container.

        Examples
        --------
        >>> import ctypes
        >>> vec = NumpyNtuples(3, 'int32').element([1, 2, 3])
        >>> arr_type = ctypes.c_int32 * 3
        >>> buffer = arr_type.from_address(vec.data_ptr)
        >>> arr = np.frombuffer(buffer, dtype='int32')
        >>> print(arr)
        [1 2 3]

        In-place modification via pointer:

        >>> arr[0] = 5
        >>> print(vec)
        [5, 2, 3]
        """
        return self.data.ctypes.data

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if all entries of other are equal to this
            vector's entries, ``False`` otherwise.

        Notes
        -----
        Space membership is not checked, hence vectors from
        different spaces can be equal.

        Examples
        --------
        >>> vec1 = NumpyNtuples(3, int).element([1, 2, 3])
        >>> vec2 = NumpyNtuples(3, int).element([-1, 2, 0])
        >>> vec1 == vec2
        False
        >>> vec2 = NumpyNtuples(3, int).element([1, 2, 3])
        >>> vec1 == vec2
        True

        Space membership matters:

        >>> vec2 = NumpyNtuples(3, float).element([1, 2, 3])
        >>> vec1 == vec2 or vec2 == vec1
        False
        """
        if other is self:
            return True
        elif other not in self.space:
            return False
        else:
            return np.array_equal(self.data, other.data)

    def copy(self):
        """Create an identical (deep) copy of this vector.

        Parameters
        ----------
        None

        Returns
        -------
        copy : `NumpyNtuplesVector`
            The deep copy

        Examples
        --------
        >>> vec1 = NumpyNtuples(3, 'int').element([1, 2, 3])
        >>> vec2 = vec1.copy()
        >>> vec2
        ntuples(3, 'int').element([1, 2, 3])
        >>> vec1 == vec2
        True
        >>> vec1 is vec2
        False
        """
        return self.space.element(self.data.copy())

    def __getitem__(self, indices):
        """Access values of this vector.

        Parameters
        ----------
        indices : int or `slice`
            The position(s) that should be accessed

        Returns
        -------
        values : scalar or `NumpyNtuplesVector`
            The value(s) at the index (indices)

        Examples
        --------
        >>> str_3 = NumpyNtuples(3, dtype='U6')  # 6-char unicode
        >>> x = str_3.element(['a', 'Hello!', '0'])
        >>> print(x[0])
        a
        >>> print(x[1:3])
        [Hello!, 0]
        >>> x[1:3].space
        ntuples(2, '<U6')
        """
        if isinstance(indices, Integral):
            return self.data[indices]  # single index
        else:
            arr = self.data[indices]
            return type(self.space)(len(arr), dtype=self.dtype).element(arr)

    def __setitem__(self, indices, values):
        """Set values of this vector.

        Parameters
        ----------
        indices : int or `slice`
            The position(s) that should be set
        values : scalar or `array-like`
            The value(s) that are to be assigned.

            If ``indices`` is an integer, ``value`` must be scalar.

            If ``indices`` is a slice, ``value`` must be
            broadcastable to the size of the slice (same size,
            shape ``(1,)`` or single value).

        Returns
        -------
        None

        Examples
        --------
        >>> int_3 = NumpyNtuples(3, 'int')
        >>> x = int_3.element([1, 2, 3])
        >>> x[0] = 5
        >>> x
        ntuples(3, 'int').element([5, 2, 3])

        Assignment from array-like structures or another
        vector:

        >>> y = NumpyNtuples(2, 'short').element([-1, 2])
        >>> x[:2] = y
        >>> x
        ntuples(3, 'int').element([-1, 2, 3])
        >>> x[1:3] = [7, 8]
        >>> x
        ntuples(3, 'int').element([-1, 7, 8])
        >>> x[:] = np.array([0, 0, 0])
        >>> x
        ntuples(3, 'int').element([0, 0, 0])

        Broadcasting is also supported:

        >>> x[1:3] = -2.
        >>> x
        ntuples(3, 'int').element([0, -2, -2])

        Array views are preserved:

        >>> y = x[::2]  # view into x
        >>> y[:] = -9
        >>> print(y)
        [-9, -9]
        >>> print(x)
        [-9, -2, -9]

        Be aware of unsafe casts and over-/underflows, there
        will be warnings at maximum.

        >>> x = NumpyNtuples(2, 'int8').element([0, 0])
        >>> maxval = 255  # maximum signed 8-bit unsigned int
        >>> x[0] = maxval + 1
        >>> x
        ntuples(2, 'int8').element([0, 0])
        """
        if isinstance(values, NumpyNtuplesVector):
            self.data[indices] = values.data
        else:
            self.data[indices] = values

    @property
    def ufunc(self):
        """`NumpyNtuplesUFuncs`, access to numpy style ufuncs.

        Examples
        --------
        >>> r2 = NumpyFn(2)
        >>> x = r2.element([1, -2])
        >>> x.ufunc.absolute()
        rn(2).element([1.0, 2.0])

        These functions can also be used with broadcasting

        >>> x.ufunc.add(3)
        rn(2).element([4.0, 1.0])

        and non-space elements

        >>> x.ufunc.subtract([3, 3])
        rn(2).element([-2.0, -5.0])

        There is also support for various reductions (sum, prod, min, max)

        >>> x.ufunc.sum()
        -1.0

        They also support an out parameter

        >>> y = r2.element([3, 4])
        >>> out = r2.element()
        >>> result = x.ufunc.add(y, out=out)
        >>> result
        rn(2).element([4.0, 2.0])
        >>> result is out
        True

        Notes
        -----
        These are optimized for use with ntuples and incur no overhead.
        """
        return NumpyNtuplesUFuncs(self)


def _blas_is_applicable(*args):
    """Whether BLAS routines can be applied or not.

    BLAS routines are available for single and double precision
    float or complex data only. If the arrays are non-contiguous,
    BLAS methods are usually slower, and array-writing routines do
    not work at all. Hence, only contiguous arrays are allowed.

    Parameters
    ----------
    x1,...,xN : `NtuplesBaseVector`
        The vectors to be tested for BLAS conformity
    """
    return (all(x.dtype == args[0].dtype and
                x.dtype in _BLAS_DTYPES and
                x.data.flags.contiguous
                for x in args))


def _lincomb(a, x1, b, x2, out, dtype):
    """Raw linear combination depending on data type."""

    # Shortcut for small problems
    if x1.size < 100:  # small array optimization
        out.data[:] = a * x1.data + b * x2.data
        return

    # Use blas for larger problems
    def fallback_axpy(x1, x2, n, a):
        """Fallback axpy implementation avoiding copy."""
        if a != 0:
            x2 /= a
            x2 += x1
            x2 *= a
        return x2

    def fallback_scal(a, x, n):
        """Fallback scal implementation."""
        x *= a
        return x

    def fallback_copy(x1, x2, n):
        """Fallback copy implementation."""
        x2[...] = x1[...]
        return x2

    if _blas_is_applicable(x1, x2, out):
        axpy, scal, copy = linalg.blas.get_blas_funcs(
            ['axpy', 'scal', 'copy'], arrays=(x1.data, x2.data, out.data))
    else:
        axpy, scal, copy = (fallback_axpy, fallback_scal, fallback_copy)

    if x1 is x2 and b != 0:
        # x1 is aligned with x2 -> out = (a+b)*x1
        _lincomb(a + b, x1, 0, x1, out, dtype)
    elif out is x1 and out is x2:
        # All the vectors are aligned -> out = (a+b)*out
        scal(a + b, out.data, native(out.size))
    elif out is x1:
        # out is aligned with x1 -> out = a*out + b*x2
        if a != 1:
            scal(a, out.data, native(out.size))
        if b != 0:
            axpy(x2.data, out.data, native(out.size), b)
    elif out is x2:
        # out is aligned with x2 -> out = a*x1 + b*out
        if b != 1:
            scal(b, out.data, native(out.size))
        if a != 0:
            axpy(x1.data, out.data, native(out.size), a)
    else:
        # We have exhausted all alignment options, so x1 != x2 != out
        # We now optimize for various values of a and b
        if b == 0:
            if a == 0:  # Zero assignment -> out = 0
                out.data[:] = 0
            else:  # Scaled copy -> out = a*x1
                copy(x1.data, out.data, native(out.size))
                if a != 1:
                    scal(a, out.data, native(out.size))
        else:
            if a == 0:  # Scaled copy -> out = b*x2
                copy(x2.data, out.data, native(out.size))
                if b != 1:
                    scal(b, out.data, native(out.size))

            elif a == 1:  # No scaling in x1 -> out = x1 + b*x2
                copy(x1.data, out.data, native(out.size))
                axpy(x2.data, out.data, native(out.size), b)
            else:  # Generic case -> out = a*x1 + b*x2
                copy(x2.data, out.data, native(out.size))
                if b != 1:
                    scal(b, out.data, native(out.size))
                axpy(x1.data, out.data, native(out.size), a)


class NumpyFn(FnBase, NumpyNtuples):

    """Vector space F^n with vector multiplication.

    This space implements n-tuples of elements from a `Field` ``F``,
    which is usually the real or complex numbers.

    Its elements are represented as instances of the `NumpyFnVector` class.
    """

    def __init__(self, size, dtype='float64', **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        size : positive int
            The number of dimensions of the space
        dtype :
            The data type of the storage array. Can be provided in any
            way the `numpy.dtype` function understands, most notably
            as built-in type, as `numpy.dtype` or as string.

            Only scalar data types are allowed.

        weight : optional
            Use weighted inner product, norm, and dist. The following
            types are supported as ``weight``:

            `FnWeightingBase`:
            Use this weighting as-is. Compatibility with this
            space's elements is not checked during init.

            float: Weighting by a constant

            array-like: Weighting by a matrix (2-dim. array) or a vector
            (1-dim. array, corresponds to a diagonal matrix). A matrix
            can also be given as a sparse matrix
            ( `scipy.sparse.spmatrix`).

            Default: no weighting

            This option cannot be combined with ``dist``,
            ``norm`` or ``inner``.

        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no
            inner product is defined.
            If ``weight`` is a sparse matrix, only 1.0, 2.0 and
            ``inf`` are allowed.

            This option is ignored if ``dist``, ``norm`` or
            ``inner`` is given.

            Default: 2.0

        Other Parameters
        ----------------

        dist : callable, optional
            The distance function defining a metric on the space.
            It must accept two `NumpyFnVector` arguments and
            fulfill the following mathematical conditions for any
            three vectors ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``

            By default, ``dist(x, y)`` is calculated as ``norm(x - y)``.
            This creates an intermediate array ``x - y``, which can be
            avoided by choosing ``dist_using_inner=True``.

            This option cannot be combined with ``weight``,
            ``norm`` or ``inner``.

        norm : callable, optional
            The norm implementation. It must accept an
            `NumpyFnVector` argument, return a float and satisfy the
            following conditions for all vectors ``x, y`` and scalars
            ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``

            By default, ``norm(x)`` is calculated as ``inner(x, x)``.

            This option cannot be combined with ``weight``,
            ``dist`` or ``inner``.

        inner : callable, optional
            The inner product implementation. It must accept two
            `NumpyFnVector` arguments, return a element from
            the field of the space (real or complex number) and
            satisfy the following conditions for all vectors
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

            This option cannot be combined with ``weight``,
            ``dist`` or ``norm``.

        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.

            Default: ``False``.

        kwargs :
            Further keyword arguments are passed to the weighting
            classes.

        See Also
        --------
        NumpyFnMatrixWeighting
        NumpyFnVectorWeighting
        NumpyFnConstWeighting

        Examples
        --------
        >>> space = NumpyFn(3, 'float')
        >>> space
        rn(3)
        >>> space = NumpyFn(3, 'float', weight=[1, 2, 3])
        >>> space
        rn(3, weight=[1, 2, 3])
        """
        # TODO: fix dead link `scipy.sparse.spmatrix`
        NumpyNtuples.__init__(self, size, dtype)
        FnBase.__init__(self, size, dtype)

        dist = kwargs.pop('dist', None)
        norm = kwargs.pop('norm', None)
        inner = kwargs.pop('inner', None)
        weight = kwargs.pop('weight', None)
        exponent = kwargs.pop('exponent', 2.0)
        dist_using_inner = bool(kwargs.pop('dist_using_inner', False))

        # Check validity of option combination (3 or 4 out of 4 must be None)
        if sum(x is None for x in (dist, norm, inner, weight)) < 3:
            raise ValueError('invalid combination of options `weight`, '
                             '`dist`, `norm` and `inner`')

        if any(x is not None for x in (dist, norm, inner)) and exponent != 2.0:
            raise ValueError('`exponent` cannot be used together with '
                             '`dist`, `norm` and `inner`')

        # Set the weighting
        if weight is not None:
            if isinstance(weight, WeightingBase):
                self._weighting = weight
            elif np.isscalar(weight):
                self._weighting = NumpyFnConstWeighting(
                    weight, exponent, dist_using_inner=dist_using_inner)
            elif weight is None:
                # Need to wait until dist, norm and inner are handled
                pass
            elif isspmatrix(weight):
                self._weighting = NumpyFnMatrixWeighting(
                    weight, exponent, dist_using_inner=dist_using_inner,
                    **kwargs)
            else:  # last possibility: make a matrix
                arr = np.asarray(weight)
                if arr.dtype == object:
                    raise ValueError('invalid weight argument {}'
                                     ''.format(weight))
                if arr.ndim == 1:
                    self._weighting = NumpyFnVectorWeighting(
                        arr, exponent, dist_using_inner=dist_using_inner)
                elif arr.ndim == 2:
                    self._weighting = NumpyFnMatrixWeighting(
                        arr, exponent, dist_using_inner=dist_using_inner,
                        **kwargs)
                else:
                    raise ValueError('array-like input {} is not 1- or '
                                     '2-dimensional'.format(weight))

        elif dist is not None:
            self._weighting = NumpyFnCustomDist(dist)
        elif norm is not None:
            self._weighting = NumpyFnCustomNorm(norm)
        elif inner is not None:
            self._weighting = NumpyFnCustomInnerProduct(inner)
        else:  # all None -> no weighing
            self._weighting = NumpyFnNoWeighting(
                exponent, dist_using_inner=dist_using_inner)

    @property
    def exponent(self):
        """Exponent of the norm and distance."""
        return self._weighting.exponent

    @property
    def weighting(self):
        """This space's weighting scheme."""
        return self._weighting

    @property
    def is_weighted(self):
        """``True`` if the weighting is not `NumpyFnNoWeighting`."""
        return not isinstance(self.weighting, NumpyFnNoWeighting)

    def _lincomb(self, a, x1, b, x2, out):
        """Linear combination of ``x1`` and ``x2``.

        Calculate ``out = a*x1 + b*x2`` using optimized BLAS
        routines if possible.

        Parameters
        ----------
        a, b : `FnBase.field` elements
            Scalars to multiply ``x1`` and ``x2`` with
        x1, x2 : `NumpyFnVector`
            Summands in the linear combination
        out : `NumpyFnVector`
            Vector to which the result is written

        Returns
        -------
        None

        Examples
        --------
        >>> c3 = NumpyFn(3, dtype=complex)
        >>> x = c3.element([1+1j, 2-1j, 3])
        >>> y = c3.element([4+0j, 5, 6+0.5j])
        >>> out = c3.element()
        >>> c3.lincomb(2j, x, 3-1j, y, out)  # out is returned
        cn(3).element([(10-2j), (17-1j), (18.5+1.5j)])
        >>> out
        cn(3).element([(10-2j), (17-1j), (18.5+1.5j)])
        """
        _lincomb(a, x1, b, x2, out, self.dtype)

    def _dist(self, x1, x2):
        """Calculate the distance between two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Vectors whose mutual distance is calculated

        Returns
        -------
        dist : float
            Distance between the vectors

        Examples
        --------
        The default case is the euclidean distance

        >>> c2_2 = NumpyFn(2, dtype=complex)
        >>> x = c2_2.element([3+1j, 4])
        >>> y = c2_2.element([1j, 4-4j])
        >>> c2_2.dist(x, y)
        5.0

        If the user has given another dist function, that one is used instead.
        For example, the 2-norm can be given explicitly:

        >>> from numpy.linalg import norm
        >>> c2_2 = NumpyFn(2, dtype=complex,
        ...                dist=lambda x, y: norm(x - y, ord=2))
        >>> x = c2_2.element([3+1j, 4])
        >>> y = c2_2.element([1j, 4-4j])
        >>> c2_2.dist(x, y)
        5.0

        Likewise, the 1-norm can be given

        >>> c2_1 = NumpyFn(2, dtype=complex,
        ...                dist=lambda x, y: norm(x - y, ord=1))
        >>> x = c2_1.element([3+1j, 4])
        >>> y = c2_1.element([1j, 4-4j])
        >>> c2_1.dist(x, y)
        7.0
        """
        return self.weighting.dist(x1, x2)

    def _norm(self, x):
        """Calculate the norm of a vector.

        Parameters
        ----------
        x : `NumpyFnVector`
            The vector whose norm is calculated

        Returns
        -------
        norm : float
            Norm of the vector

        Examples
        --------
        >>> from numpy.linalg import norm

        2-norm

        >>> c2_2 =  NumpyFn(2, dtype=complex, norm=norm)
        >>> x = c2_2.element([3+1j, 1-5j])
        >>> c2_2.norm(x)
        6.0

        1-norm

        >>> from functools import partial
        >>> c2_1 = NumpyFn(2, dtype=complex, norm=partial(norm, ord=1))
        >>> x = c2_1.element([3-4j, 12+5j])
        >>> c2_1.norm(x)
        18.0
        """
        return self.weighting.norm(x)

    def _inner(self, x1, x2):
        """Raw inner product of two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            The vectors whose inner product is calculated

        Returns
        -------
        inner : `field` element
            Inner product of the vectors

        Examples
        --------
        >>> c3 = NumpyFn(2, dtype=complex, inner=lambda x, y: np.vdot(y, x))
        >>> x = c3.element([5+1j, -2j])
        >>> y = c3.element([1, 1+1j])
        >>> c3.inner(x, y) == (5+1j)*1 + (-2j)*(1-1j)
        True

        Define a space with custom inner product:

        >>> weights = np.array([1., 2.])
        >>> def weighted_inner(x, y):
        ...     return np.vdot(weights * y.data, x.data)

        >>> c3w = NumpyFn(2, dtype=complex, inner=weighted_inner)
        >>> x = c3w.element(x)  # elements must be cast (no copy)
        >>> y = c3w.element(y)
        >>> c3w.inner(x, y) == 1*(5+1j)*1 + 2*(-2j)*(1-1j)
        True
        """
        return self.weighting.inner(x1, x2)

    def _multiply(self, x1, x2, out):
        """Entry-wise product of two vectors, assigned to out.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Factors in the product
        out : `NumpyFnVector`
            Vector to which the result is written

        Returns
        -------
        None

        Examples
        --------
        >>> c3 = NumpyFn(3, dtype=complex)
        >>> x = c3.element([5+1j, 3, 2-2j])
        >>> y = c3.element([1, 2+1j, 3-1j])
        >>> out = c3.element()
        >>> c3.multiply(x, y, out)  # out is returned
        cn(3).element([(5+1j), (6+3j), (4-8j)])
        >>> out
        cn(3).element([(5+1j), (6+3j), (4-8j)])
        """
        np.multiply(x1.data, x2.data, out=out.data)

    def _divide(self, x1, x2, out):
        """Entry-wise division of two vectors, assigned to out.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Dividend and divisor in the quotient
        out : `NumpyFnVector`
            Vector to which the result is written

        Returns
        -------
        None

        Examples
        --------
        >>> r3 = NumpyFn(3)
        >>> x = r3.element([3, 5, 6])
        >>> y = r3.element([1, 2, 2])
        >>> out = r3.element()
        >>> r3.divide(x, y, out)  # out is returned
        rn(3).element([3.0, 2.5, 3.0])
        >>> out
        rn(3).element([3.0, 2.5, 3.0])
        """
        np.divide(x1.data, x2.data, out=out.data)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is an instance of this space's type
            with the same `NtuplesBase.size` and `NtuplesBase.dtype`,
            and identical distance function, ``False`` otherwise.

        Examples
        --------
        >>> from numpy.linalg import norm
        >>> def dist(x, y, ord):
        ...     return norm(x - y, ord)

        >>> from functools import partial
        >>> dist2 = partial(dist, ord=2)
        >>> c3 = NumpyFn(3, dtype=complex, dist=dist2)
        >>> c3_same = NumpyFn(3, dtype=complex, dist=dist2)
        >>> c3  == c3_same
        True

        Different ``dist`` functions result in different spaces - the
        same applies for ``norm`` and ``inner``:

        >>> dist1 = partial(dist, ord=1)
        >>> r3_1 = NumpyFn(3, dist=dist1)
        >>> r3_2 = NumpyFn(3, dist=dist2)
        >>> r3_1 == r3_2
        False

        Be careful with Lambdas - they result in non-identical function
        objects:

        >>> r3_lambda1 = NumpyFn(3, dist=lambda x, y: norm(x-y, ord=1))
        >>> r3_lambda2 = NumpyFn(3, dist=lambda x, y: norm(x-y, ord=1))
        >>> r3_lambda1 == r3_lambda2
        False

        An `NumpyFn` space with the same data type is considered equal:

        >>> c3 = NumpyFn(3, dtype=complex)
        >>> f3_cdouble = NumpyFn(3, dtype='complex128')
        >>> c3 == f3_cdouble
        True
        """
        if other is self:
            return True

        return (NumpyNtuples.__eq__(self, other) and
                self.weighting == other.weighting)

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.is_rn:
            constructor_name = 'rn'
        elif self.is_cn:
            constructor_name = 'cn'
        else:
            constructor_name = 'fn'

        inner_str = '{}'.format(self.size)
        if self.dtype != self.default_dtype(self.field):
            inner_str += ', {}'.format(dtype_repr(self.dtype))

        weight_str = self.weighting.repr_part
        if weight_str:
            inner_str += ', ' + weight_str
        return '{}({})'.format(constructor_name, inner_str)

    # Copy these to handle bug in ABCmeta
    zero = NumpyNtuples.zero
    one = NumpyNtuples.one

    @property
    def element_type(self):
        """`NumpyFnVector`"""
        return NumpyFnVector

    @staticmethod
    def available_dtypes():
        """Available data types.

        Notes
        -----
        This is the set of all arithmetic dtypes available to numpy. See
        `numpy.sctypes` for more information.

        The available dtypes may depend on the specific system used.
        """
        # TODO: fix dead link `numpy.sctypes`
        return np.sctypes['int'] + np.sctypes['float'] + np.sctypes['complex']

    @staticmethod
    def default_dtype(field=None):
        """Return the default of `NumpyFn` data type for a given field.

        Parameters
        ----------
        field : `Field`, optional
            Set of numbers to be represented by a data type.
            Currently supported : `RealNumbers`, `ComplexNumbers`.
            Default: `RealNumbers`

        Returns
        -------
        dtype : `numpy.dtype`
            Numpy data type specifier. The returned defaults are:

            ``RealNumbers()`` : ``np.dtype('float64')``

            ``ComplexNumbers()`` : ``np.dtype('complex128')``
        """
        if field is None or field == RealNumbers():
            return np.dtype('float64')
        elif field == ComplexNumbers():
            return np.dtype('complex128')
        else:
            raise ValueError('no default data type defined for field {}.'
                             ''.format(field))


class NumpyFnVector(FnBaseVector, NumpyNtuplesVector):

    """Representation of an `NumpyFn` element."""

    def __init__(self, space, data):
        """Initialize a new instance."""
        if not isinstance(space, NumpyFn):
            raise TypeError('{!r} not an `NumpyFn` instance'
                            ''.format(space))

        FnBaseVector.__init__(self, space)
        NumpyNtuplesVector.__init__(self, space, data)

    @property
    def real(self):
        """Real part of this vector.

        Returns
        -------
        real : `NumpyFnVector` view with real dtype
            The real part this vector as a vector in `rn`

        Examples
        --------
        >>> c3 = NumpyFn(3, dtype=complex)
        >>> x = c3.element([5+1j, 3, 2-2j])
        >>> x.real
        rn(3).element([5.0, 3.0, 2.0])

        The `rn` vector is really a view, so changes affect
        the original array:

        >>> x.real *= 2
        >>> x
        cn(3).element([(10+1j), (6+0j), (4-2j)])
        """
        real_space = NumpyFn(self.space.size, self.space.real_dtype)
        return real_space.element(self.data.real)

    @real.setter
    def real(self, newreal):
        """Setter for the real part.

        This method is invoked by ``vec.real = other``.

        Parameters
        ----------
        newreal : `array-like` or scalar
            The new real part for this vector.

        Examples
        --------
        >>> c3 = NumpyFn(3, dtype=complex)
        >>> x = c3.element([5+1j, 3, 2-2j])
        >>> a = NumpyFn(3).element([0, 0, 0])
        >>> x.real = a
        >>> x
        cn(3).element([1j, 0j, -2j])

        Other array-like types and broadcasting:

        >>> x.real = 1.0
        >>> x
        cn(3).element([(1+1j), (1+0j), (1-2j)])
        >>> x.real = [0, 2, -1]
        >>> x
        cn(3).element([1j, (2+0j), (-1-2j)])
        """
        self.real.data[:] = newreal

    @property
    def imag(self):
        """Imaginary part of this vector.

        Returns
        -------
        imag : `NumpyFnVector`
            The imaginary part this vector as a vector in `rn`

        Examples
        --------
        >>> c3 = NumpyFn(3, dtype=complex)
        >>> x = c3.element([5+1j, 3, 2-2j])
        >>> x.imag
        rn(3).element([1.0, 0.0, -2.0])

        The `rn` vector is really a view, so changes affect
        the original array:

        >>> x.imag *= 2
        >>> x
        cn(3).element([(5+2j), (3+0j), (2-4j)])
        """
        real_space = NumpyFn(self.space.size, self.space.real_dtype)
        return real_space.element(self.data.imag)

    @imag.setter
    def imag(self, newimag):
        """Setter for the imaginary part.

        This method is invoked by ``vec.imag = other``.

        Parameters
        ----------
        newimag : `array-like` or scalar
            The new imaginary part for this vector.

        Examples
        --------
        >>> x = cn(3).element([5+1j, 3, 2-2j])
        >>> a = NumpyFn(3).element([0, 0, 0])
        >>> x.imag = a; print(x)
        [(5+0j), (3+0j), (2+0j)]

        Other array-like types and broadcasting:

        >>> x.imag = 1.0; print(x)
        [(5+1j), (3+1j), (2+1j)]
        >>> x.imag = [0, 2, -1]; print(x)
        [(5+0j), (3+2j), (2-1j)]
        """
        self.imag.data[:] = newimag

    def conj(self, out=None):
        """Complex conjugate of this vector.

        Parameters
        ----------
        out : `NumpyFnVector`, optional
            Vector to which the complex conjugate is written.
            Must be an element of this vector's space.

        Returns
        -------
        out : `NumpyFnVector`
            The complex conjugate vector. If ``out`` was provided,
            the returned object is a reference to it.

        Examples
        --------
        >>> x = NumpyFn(3, dtype=complex).element([5+1j, 3, 2-2j])
        >>> y = x.conj(); print(y)
        [(5-1j), (3-0j), (2+2j)]

        The out parameter allows you to avoid a copy

        >>> z = NumpyFn(3, dtype=complex).element()
        >>> z_out = x.conj(out=z); print(z)
        [(5-1j), (3-0j), (2+2j)]
        >>> z_out is z
        True

        It can also be used for in-place conj

        >>> x_out = x.conj(out=x); print(x)
        [(5-1j), (3-0j), (2+2j)]
        >>> x_out is x
        True
        """
        if out is None:
            return self.space.element(self.data.conj())
        else:
            self.data.conj(out.data)
            return out

    def __ipow__(self, other):
        """Return ``self **= other``."""
        try:
            if other == int(other):
                return super().__ipow__(other)
        except TypeError:
            pass

        np.power(self.data, other, out=self.data)
        return self


class MatVecOperator(Operator):

    """Matrix multiply operator :math:`\mathbb{F}^n -> \mathbb{F}^m`.

    This operator uses a matrix to represent an operator, and get its adjoint
    and inverse by doing computations on the matrix. This is in general a
    rather slow approach, and users are recommended to use other alternatives
    if available.
    """

    def __init__(self, matrix, domain=None, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        matrix : `array-like` or  `scipy.sparse.spmatrix`
            Matrix representing the linear operator. Its shape must be
            ``(m, n)``, where ``n`` is the size of ``domain`` and ``m`` the
            size of ``range``. Its dtype must be castable to the range
            ``dtype``.
        domain : `NumpyFn`, optional
            Space on whose elements the matrix acts. If not provided,
            the domain is inferred from the matrix ``dtype`` and
            ``shape``. If provided, its dtype must be castable to the
            range dtype.
        range : `NumpyFn`, optional
            Space to which the matrix maps. If not provided,
            the domain is inferred from the matrix ``dtype`` and
            ``shape``.
        """
        # TODO: fix dead link `scipy.sparse.spmatrix`
        if isspmatrix(matrix):
            self.__matrix = matrix
        else:
            self.__matrix = np.asarray(matrix)

        if self.matrix.ndim != 2:
            raise ValueError('matrix {} has {} axes instead of 2'
                             ''.format(matrix, self.matrix.ndim))

        # Infer domain and range from matrix if necessary
        if domain is None:
            domain = NumpyFn(self.matrix.shape[1], dtype=self.matrix.dtype)
        elif not isinstance(domain, NumpyFn):
            raise TypeError('`domain` {!r} is not an `NumpyFn` instance'
                            ''.format(domain))

        if range is None:
            range = NumpyFn(self.matrix.shape[0], dtype=self.matrix.dtype)
        elif not isinstance(range, NumpyFn):
            raise TypeError('`range` {!r} is not an `NumpyFn` instance'
                            ''.format(range))

        # Check compatibility of matrix with domain and range
        if not np.can_cast(domain.dtype, range.dtype):
            raise TypeError('domain data type {!r} cannot be safely cast to '
                            'range data type {!r}'
                            ''.format(domain.dtype, range.dtype))

        if self.matrix.shape != (range.size, domain.size):
            raise ValueError('matrix shape {} does not match the required '
                             'shape {} of a matrix {} --> {}'
                             ''.format(self.matrix.shape,
                                       (range.size, domain.size),
                                       domain, range))
        if not np.can_cast(self.matrix.dtype, range.dtype):
            raise TypeError('matrix data type {!r} cannot be safely cast to '
                            'range data type {!r}.'
                            ''.format(matrix.dtype, range.dtype))

        super().__init__(domain, range, linear=True)

    @property
    def matrix(self):
        """Matrix representing this operator."""
        return self.__matrix

    @property
    def matrix_issparse(self):
        """Whether the representing matrix is sparse or not."""
        return isspmatrix(self.matrix)

    @property
    def adjoint(self):
        """Adjoint operator represented by the adjoint matrix.

        Returns
        -------
        adjoint : `MatVecOperator`
        """
        if self.domain.field != self.range.field:
            raise NotImplementedError('adjoint not defined since fields '
                                      'of domain and range differ ({} != {})'
                                      ''.format(self.domain.field,
                                                self.range.field))
        return MatVecOperator(self.matrix.conj().T,
                              domain=self.range, range=self.domain)

    @property
    def inverse(self):
        """Inverse operator represented by the inverse matrix.

        Taking the inverse causes sparse matrices to become dense and is
        generally very heavy computationally since the matrix is inverted
        numerically (an O(n^3) operation). It is recommended to instead
        use one of the solvers available in the ``odl.solvers`` package.

        Returns
        -------
        inverse : `MatVecOperator`
        """
        if self.matrix_issparse:
            dense_matrix = self.matrix.toarray()
        else:
            dense_matrix = self.matrix

        return MatVecOperator(np.linalg.inv(dense_matrix),
                              domain=self.range, range=self.domain)

    def _call(self, x, out=None):
        """Raw apply method on input, writing to given output."""
        if out is None:
            return self.range.element(self.matrix.dot(x.data))
        else:
            if self.matrix_issparse:
                # Unfortunately, there is no native in-place dot product for
                # sparse matrices
                out.data[:] = self.matrix.dot(x.data)
            else:
                self.matrix.dot(x.data, out=out.data)

    # TODO: repr and str


def _weighting(weight, exponent, dist_using_inner=False):
    """Return a weighting whose type is inferred from the arguments."""
    if np.isscalar(weight):
        weighting = NumpyFnConstWeighting(
            weight, exponent=exponent, dist_using_inner=dist_using_inner)
    elif isspmatrix(weight):
        weighting = NumpyFnMatrixWeighting(
            weight, exponent=exponent, dist_using_inner=dist_using_inner)
    else:
        weight_ = np.asarray(weight)
        if weight_.dtype == object:
            raise ValueError('bad weight {}'.format(weight))
        if weight_.ndim == 1:
            weighting = NumpyFnVectorWeighting(
                weight_, exponent=exponent, dist_using_inner=dist_using_inner)
        elif weight_.ndim == 2:
            weighting = NumpyFnMatrixWeighting(
                weight_, exponent=exponent, dist_using_inner=dist_using_inner)
        else:
            raise ValueError('array-like weight must have 1 or 2 dimensions, '
                             'but {} has {} dimensions'
                             ''.format(weight, weight_.ndim))
    return weighting


def npy_weighted_inner(weight):
    """Weighted inner product on `NumpyFn` spaces as free function.

    Parameters
    ----------
    weight : scalar or `array-like`
        Weight of the inner product. A scalar is interpreted as a
        constant weight, a 1-dim. array as a weighting vector and a
        2-dimensional array as a weighting matrix.

    Returns
    -------
    inner : callable
        Inner product function with given weight. Constant weightings
        are applicable to spaces of any size, for arrays the sizes
        of the weighting and the space must match.

    See Also
    --------
    NumpyFnConstWeighting, NumpyFnVectorWeighting, NumpyFnMatrixWeighting
    """
    return _weighting(weight, exponent=2.0).inner


def npy_weighted_norm(weight, exponent=2.0):
    """Weighted norm on `NumpyFn` spaces as free function.

    Parameters
    ----------
    weight : scalar or `array-like`
        Weight of the norm. A scalar is interpreted as a
        constant weight, a 1-dim. array as a weighting vector and a
        2-dimensional array as a weighting matrix.
    exponent : positive float
        Exponent of the norm. If ``weight`` is a sparse matrix, only
        1.0, 2.0 and ``inf`` are allowed.

    Returns
    -------
    norm : callable
        Norm function with given weight. Constant weightings
        are applicable to spaces of any size, for arrays the sizes
        of the weighting and the space must match.

    See Also
    --------
    NumpyFnConstWeighting, NumpyFnVectorWeighting, NumpyFnMatrixWeighting
    """
    return _weighting(weight, exponent=exponent).norm


def npy_weighted_dist(weight, exponent=2.0, use_inner=False):
    """Weighted distance on `NumpyFn` spaces as free function.

    Parameters
    ----------
    weight : scalar or `array-like`
        Weight of the distance. A scalar is interpreted as a
        constant weight, a 1-dim. array as a weighting vector and a
        2-dimensional array as a weighting matrix.
    exponent : positive float
        Exponent of the norm. If ``weight`` is a sparse matrix, only
        1.0, 2.0 and ``inf`` are allowed.
    use_inner : bool, optional
        Calculate ``dist`` using the formula

            ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

        This avoids the creation of new arrays and is thus faster
        for large arrays. On the downside, it will not evaluate to
        exactly zero for equal (but not identical) ``x`` and ``y``.

        Can only be used if ``exponent`` is 2.0.

    Returns
    -------
    dist : callable
        Distance function with given weight. Constant weightings
        are applicable to spaces of any size, for arrays the sizes
        of the weighting and the space must match.

    See Also
    --------
    NumpyFnConstWeighting, NumpyFnVectorWeighting, NumpyFnMatrixWeighting
    """
    return _weighting(weight, exponent=exponent,
                      dist_using_inner=use_inner).dist


def _norm_default(x):
    """Default Euclidean norm implementation."""
    if _blas_is_applicable(x):
        nrm2 = linalg.blas.get_blas_funcs('nrm2', dtype=x.dtype)
        norm = partial(nrm2, n=native(x.size))
    else:
        norm = np.linalg.norm
    return norm(x.data)


def _pnorm_default(x, p):
    """Default p-norm implementation."""
    return np.linalg.norm(x.data, ord=p)


def _pnorm_diagweight(x, p, w):
    """Diagonally weighted p-norm implementation."""
    # This is faster than first applying the weights and then summing with
    # BLAS dot or nrm2
    xp = np.abs(x.data)
    if np.isfinite(p):
        xp = np.power(xp, p, out=xp)
        xp *= w  # w is a plain NumPy array
        return np.sum(xp) ** (1 / p)
    else:
        xp *= w
        return np.max(xp)


def _inner_default(x1, x2):
    """Default Euclidean inner product implementation."""
    if _blas_is_applicable(x1, x2):
        dotc = linalg.blas.get_blas_funcs('dotc', dtype=x1.dtype)
        dot = partial(dotc, n=native(x1.size))
    elif is_real_dtype(x1.dtype):
        dot = np.dot  # still much faster than vdot
    else:
        dot = np.vdot  # slowest alternative
    # x2 as first argument because we want linearity in x1
    return dot(x2.data, x1.data)


class NumpyFnMatrixWeighting(MatrixWeightingBase):

    """Matrix weighting for `NumpyFn`.

    For exponent 2.0, a new weighted inner product with matrix ``W``
    is defined as::

        <a, b>_W := <W * a, b> = b^H * W * a

    with ``b^H`` standing for transposed complex conjugate.

    For other exponents, only norm and dist are defined. In the case of
    exponent ``inf``, the weighted norm is::

        ||a||_{W, inf} := ||W * a||_inf

    otherwise it is::

        ||a||_{W, p} := ||W^{1/p} * a||_p

    Note that this definition does **not** fulfill the limit property
    in ``p``, i.e.::

        ||x||_{W, p} --/-> ||x||_{W, inf}  for p --> inf

    unless ``W`` is the identity matrix.

    The matrix must be Hermitian and posivive definite, otherwise it
    does not define an inner product or norm, respectively. This is not
    checked during initialization.
    """

    def __init__(self, matrix, exponent=2.0, dist_using_inner=False, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        matrix :  `scipy.sparse.spmatrix` or `array-like`, 2-dim.
            Square weighting matrix of the inner product
        exponent : positive float
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
            If ``matrix`` is a sparse matrix, only 1.0, 2.0 and ``inf``
            are allowed.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.
        precomp_mat_pow : bool, optional
            If ``True``, precompute the matrix power ``W ** (1/p)``
            during initialization. This has no effect if ``exponent``
            is 1.0, 2.0 or ``inf``.

            Default: False

        cache_mat_pow : bool, optional
            If ``True``, cache the matrix power ``W ** (1/p)``. This can
            happen either during initialization or in the first call to
            ``norm`` or ``dist``, resp. This has no effect if
            ``exponent`` is 1.0, 2.0 or ``inf``.

            Default: True

        cache_mat_decomp : bool, optional
            If ``True``, cache the eigenbasis decomposition of the
            matrix. This can happen either during initialization or in
            the first call to ``norm`` or ``dist``, resp. This has no
            effect if ``exponent`` is 1.0, 2.0 or ``inf``.

            Default: False

        Notes
        -----
        The matrix power ``W ** (1/p)`` is computed with by eigenbasis
        decomposition::

            eigval, eigvec = scipy.linalg.eigh(matrix)
            mat_pow = (eigval ** p * eigvec).dot(eigvec.conj().T)

        Depending on the matrix size, this can be rather expensive.
        """
        # TODO: fix dead link `scipy.sparse.spmatrix`
        super().__init__(matrix, impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner, **kwargs)

    def inner(self, x1, x2):
        """Calculate the matrix-weighted inner product of two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Vectors whose inner product is calculated

        Returns
        -------
        inner : float or complex
            The inner product of the vectors
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))
        else:
            inner = _inner_default(x1.space.element(self.matrix.dot(x1)), x2)
            if is_real_dtype(x1.dtype):
                return float(inner)
            else:
                return complex(inner)

    def norm(self, x):
        """Calculate the matrix-weighted norm of a vector.

        Parameters
        ----------
        x : `NumpyFnVector`
            Vector whose norm is calculated

        Returns
        -------
        norm : float
            The norm of the vector
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?
            return np.sqrt(norm_squared)

        if self._mat_pow is None:
            # This case can only be reached if p != 1,2,inf
            if self.matrix_issparse:
                raise NotImplementedError('sparse matrix powers not '
                                          'suppoerted')

            if self._eigval is None or self._eigvec is None:
                # No cached decomposition, computing new one
                eigval, eigvec = linalg.eigh(self.matrix)
                if self._cache_mat_decomp:
                    self._eigval, self._eigvec = eigval, eigvec
                    eigval_pow = eigval ** (1.0 / self.exponent)
                else:
                    # Not storing eigenvalues, so we can destroy them
                    eigval_pow = eigval
                    eigval_pow **= 1.0 / self.exponent
            else:
                # Using cached, cannot destroy
                eigval, eigvec = self._eigval, self._eigvec
                eigval_pow = eigval ** (1.0 / self.exponent)

            mat_pow = (eigval_pow * eigvec).dot(eigvec.conj().T)
            if self._cache_mat_pow:
                self._mat_pow = mat_pow
        else:
            mat_pow = self._mat_pow

        return float(_pnorm_default(x.space.element(mat_pow.dot(x)),
                                    self.exponent))


class NumpyFnVectorWeighting(VectorWeightingBase):

    """Vector weighting for `NumpyFn`.

    For exponent 2.0, a new weighted inner product with vector ``w``
    is defined as::

        <a, b>_w := <w * a, b> = b^H (w * a)

    with ``b^H`` standing for transposed complex conjugate and
    ``w * a`` for entry-wise multiplication.

    For other exponents, only norm and dist are defined. In the case of
    exponent ``inf``, the weighted norm is

        ||a||_{w, inf} := ||w * a||_inf

    otherwise it is::

        ||a||_{w, p} := ||w^{1/p} * a||

    Note that this definition does **not** fulfill the limit property
    in ``p``, i.e.::

        ||x||_{w, p} --/-> ||x||_{w, inf}  for p --> inf

    unless ``w = (1,...,1)``.

    The vector may only have positive entries, otherwise it does not
    define an inner product or norm, respectively. This is not checked
    during initialization.
    """

    def __init__(self, vector, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        vector : `array-like`, one-dim.
            Weighting vector of the inner product, norm and distance
        exponent : positive float
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.
        """
        super().__init__(vector, impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)

    def inner(self, x1, x2):
        """Calculate the vector weighted inner product of two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Vectors whose inner product is calculated

        Returns
        -------
        inner : float or complex
            The inner product of the two provided vectors
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))
        else:
            inner = _inner_default(x1 * self.vector, x2)
            if is_real_dtype(x1.dtype):
                return float(inner)
            else:
                return complex(inner)

    def norm(self, x):
        """Calculate the vector-weighted norm of a vector.

        Parameters
        ----------
        x : `NumpyFnVector`
            Vector whose norm is calculated

        Returns
        -------
        norm : float
            The norm of the provided vector
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            if norm_squared < 0:
                norm_squared = 0.0  # Compensate for numerical error
            return np.sqrt(norm_squared)
        else:
            return float(_pnorm_diagweight(x, self.exponent, self.vector))


class NumpyFnConstWeighting(ConstWeightingBase):

    """Weighting of `NumpyFn` by a constant.

    For exponent 2.0, a new weighted inner product with constant
    ``c`` is defined as::

        <a, b>_c = c * <a, b> = c * b^H a

    with ``b^H`` standing for transposed complex conjugate.

    For other exponents, only norm and dist are defined. In the case of
    exponent ``inf``, the weighted norm is defined as::

        ||a||_{c, inf} := c ||a||_inf

    otherwise it is::

        ||a||_{c, p} := c^{1/p}  ||a||_p

    Note that this definition does **not** fulfill the limit property
    in ``p``, i.e.::

        ||a||_{c,p} --/-> ||a||_{c,inf}  for p --> inf

    unless ``c = 1``.

    The constant must be positive, otherwise it does not define an
    inner product or norm, respectively.
    """

    def __init__(self, constant, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        constant : positive float
            Weighting constant of the inner product.
        exponent : positive float
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.
        """
        super().__init__(constant, impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)

    def inner(self, x1, x2):
        """Calculate the constant-weighted inner product of two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Vectors whose inner product is calculated

        Returns
        -------
        inner : float or complex
            The inner product of the two provided vectors
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))
        else:
            inner = self.const * _inner_default(x1, x2)
            return x1.space.field.element(inner)

    def norm(self, x):
        """Calculate the constant-weighted norm of a vector.

        Parameters
        ----------
        x1 : `NumpyFnVector`
            Vector whose norm is calculated

        Returns
        -------
        norm : float
            The norm of the vector
        """
        if self.exponent == 2.0:
            return np.sqrt(self.const) * float(_norm_default(x))
        elif self.exponent == float('inf'):
            return self.const * float(_pnorm_default(x, self.exponent))
        else:
            return (self.const ** (1 / self.exponent) *
                    float(_pnorm_default(x, self.exponent)))

    def dist(self, x1, x2):
        """Calculate the constant-weighted distance between two vectors.

        Parameters
        ----------
        x1, x2 : `NumpyFnVector`
            Vectors whose mutual distance is calculated

        Returns
        -------
        dist : float
            The distance between the vectors
        """
        if self.dist_using_inner:
            dist_squared = (_norm_default(x1) ** 2 + _norm_default(x2) ** 2 -
                            2 * _inner_default(x1, x2).real)
            if dist_squared < 0.0:  # Compensate for numerical error
                dist_squared = 0.0
            return np.sqrt(self.const) * float(np.sqrt(dist_squared))
        elif self.exponent == 2.0:
            return np.sqrt(self.const) * _norm_default(x1 - x2)
        elif self.exponent == float('inf'):
            return self.const * float(_pnorm_default(x1 - x2, self.exponent))
        else:
            return (self.const ** (1 / self.exponent) *
                    float(_pnorm_default(x1 - x2, self.exponent)))


class NumpyFnNoWeighting(NoWeightingBase, NumpyFnConstWeighting):

    """Weighting of `NumpyFn` with constant 1.

    For exponent 2.0, the unweighted inner product is defined as::

        <a, b> := b^H a

    with ``b^H`` standing for transposed complex conjugate.

    For other exponents, only norm and dist are defined.
    """

    # Implement singleton pattern for efficiency in the default case
    _instance = None

    def __new__(cls, *args, **kwargs):
        """Implement singleton pattern if ``exp==2.0``."""
        if len(args) == 0:
            exponent = kwargs.pop('exponent', 2.0)
            dist_using_inner = kwargs.pop('dist_using_inner', False)
        elif len(args) == 1:
            exponent = args[0]
            args = args[1:]
            dist_using_inner = kwargs.pop('dist_using_inner', False)
        else:
            exponent = args[0]
            dist_using_inner = args[1]
            args = args[2:]

        if exponent == 2.0 and not dist_using_inner:
            if not cls._instance:
                cls._instance = super().__new__(cls, *args, **kwargs)
            return cls._instance
        else:
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        exponent : positive float
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.
        """
        super().__init__(impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)


class NumpyFnCustomInnerProduct(CustomInnerProductBase):

    """Class for handling a user-specified inner product in `NumpyFn`."""

    def __init__(self, inner, dist_using_inner=True):
        """Initialize a new instance.

        Parameters
        ----------
        inner : callable
            The inner product implementation. It must accept two
            `NumpyFnVector` arguments, return an element from their space's
            field (real or complex number) and satisfy the following
            conditions for all vectors ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.
        """
        super().__init__(inner, impl='numpy',
                         dist_using_inner=dist_using_inner)


class NumpyFnCustomNorm(CustomNormBase):

    """Class for handling a user-specified norm in `NumpyFn`.

    Note that this removes ``inner``.
    """

    def __init__(self, norm):
        """Initialize a new instance.

        Parameters
        ----------
        norm : callable
            The norm implementation. It must accept an `NumpyFnVector`
            argument, return a float and satisfy the following
            conditions for all vectors ``x, y`` and scalars ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``
        """
        super().__init__(norm, impl='numpy')


class NumpyFnCustomDist(CustomDistBase):

    """Class for handling a user-specified distance in `NumpyFn`.

    Note that this removes ``inner`` and ``norm``.
    """

    def __init__(self, dist):
        """Initialize a new instance.

        Parameters
        ----------
        dist : callable
            The distance function defining a metric on `NumpyFn`. It must
            accept two `NumpyFnVector` arguments, return a float and
            fulfill the following mathematical conditions for any three
            vectors ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``
        """
        super().__init__(dist, impl='numpy')


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Gradient-based optimization schemes."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()


import numpy as np
from odl.solvers.util import ConstantLineSearch


__all__ = ('steepest_descent',)


# TODO: update all docs


def steepest_descent(f, x, line_search=1.0, maxiter=1000, tol=1e-16,
                     projection=None, callback=None):
    """Steepest descent method to minimize an objective function.

    General implementation of steepest decent (also known as gradient
    decent) for solving

        :math:`\min f(x)`

    The algorithm is intended for unconstrained problems. It needs line
    search in order guarantee convergence. With appropriate line search,
    it can also be used for constrained problems where one wants to
    minimize over some given set :math:`C`. This can be done by defining
    :math:`f(x) = \infty` for :math:`x\\not\\in C`, or by providing a
    ``projection`` function that projects the iterates on :math:`C`.

    The algorithm is described in [BV2004]_, section 9.3--9.4
    (`book available online
    <http://stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf>`_),
    [GNS2009]_, Section 12.2, and wikipedia
    `Gradient_descent
    <https://en.wikipedia.org/wiki/Gradient_descent>`_.

    Parameters
    ----------
    f : `Functional`
        Goal functional. Needs to have ``f.gradient``.
    x : ``f.domain`` element
        Starting point of the iteration
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, uses it as a
        fixed step length.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance that should be used for terminating the iteration.
    projection : callable, optional
        Function that can be used to modify the iterates in each iteration,
        for example enforcing positivity. The function should take one
        argument and modify it in-place.
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    See Also
    --------
    odl.solvers.iterative.iterative.landweber :
        Optimized solver for the case ``f(x) = ||Ax - b||_2^2``
    odl.solvers.iterative.iterative.conjugate_gradient :
        Optimized solver for the case ``f(x) = x^T Ax - 2 x^T b``
    """
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `grad` {!r}'
                        ''.format(x, grad.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    grad_x = grad.range.element()
    for _ in range(maxiter):
        grad(x, out=grad_x)

        dir_derivative = -grad_x.norm() ** 2
        if np.abs(dir_derivative) < tol:
            return  # we have converged
        step = line_search(x, -grad_x, dir_derivative)

        x.lincomb(1, x, -step, grad_x)

        if projection is not None:
            projection(x)

        if callback is not None:
            callback(x)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Maximum Likelihood Expectation Maximization algorithm."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np

__all__ = ('mlem', 'loglikelihood')


AVAILABLE_MLEM_NOISE = ('poisson',)


def mlem(op, x, data, niter=1, noise='poisson', callback=None):

    """Maximum Likelihood Expectation Maximation algorithm.

    Attempts to solve::

        max_x L(data | x)

    where ``L(data, | x)`` is the likelihood of ``data`` given ``x``. The
    likelihood depends on the forward operator ``op`` such that
    (approximately)::

        op(x) = data

    With the precise form of *approximately* is determined by ``noise``.

    Parameters
    ----------
    op : `Operator`
        Forward operator in the inverse problem.
    x : ``op.domain`` element
        Vector to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    data : ``op.range`` element
        Right-hand side of the equation defining the inverse problem.
    niter : int, optional
        Number of iterations.
    noise : {'poisson'}, optional
        Noise model determining the variant of MLEM.
        For ``'poisson'``, the initial value of ``x`` should be
        non-negative.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    Given a forward model :math:`A`, data :math:`g` and noise model :math:`X`,
    the algorithm attempts find an :math:`x` that maximizes:

    .. math::

        P(g | g \\text{ is } X(A(x)) \\text{ distributed})

    where the expectation of :math:`X(A(x))` satisfies

    .. math::

        \\mathbb{E}(X(A(x))) = A(x)

    with 'poisson' noise the algorithm is given by:

    .. math::

       x_{n+1} = \\frac{x_n}{A^* 1} A^* (g / A(x_n))

    See Also
    --------
    loglikelihood : Function for calculating the logarithm of the likelihood
    """
    noise, noise_in = str(noise).lower(), noise
    if noise not in AVAILABLE_MLEM_NOISE:
        raise NotImplemented("noise '{}' not understood"
                             ''.format(noise_in))

    if noise == 'poisson':
        if np.any(np.less(x, 0)):
            raise ValueError('`x` must be non-negative')

        eps = 1e-8

        norm = np.maximum(op.adjoint(op.range.one()), eps)
        tmp_dom = op.domain.element()
        tmp_ran = op.range.element()

        for _ in range(niter):
            op(x, out=tmp_ran)
            tmp_ran.ufunc.maximum(eps, out=tmp_ran)
            data.divide(tmp_ran, out=tmp_ran)

            op.adjoint(tmp_ran, out=tmp_dom)
            tmp_dom /= norm

            x *= tmp_dom

            if callback is not None:
                callback(x)
    else:
        raise RuntimeError('unknown noise model')


def loglikelihood(x, data, noise='poisson'):
    """log-likelihood of ``data`` given noise parametrized by ``x``.

    Parameters
    ----------
    x : ``op.domain`` element
        Value to condition the log-likelihood on.
    data : ``op.range`` element
        Data whose log-likelihood given ``x`` shall be calculated.
    noise : {'poisson'}, optional
        The type of noise.
    """
    noise, noise_in = str(noise).lower(), noise
    if noise not in AVAILABLE_MLEM_NOISE:
        raise NotImplemented("noise '{}' not understood"
                             ''.format(noise_in))

    if noise == 'poisson':
        if np.any(np.less(x, 0)):
            raise ValueError('`x` must be non-negative')

        return np.sum(data * np.log(x + 1e-8) - x)
    else:
        raise RuntimeError('unknown noise model')



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Bindings to the PyWavelets backend for wavelet transforms.

`PyWavelets <https://pywavelets.readthedocs.io/>`_ is a Python library
for wavelet transforms in arbitrary dimensions, featuring a large number
of built-in wavelet filters.
"""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

from itertools import product
import numpy as np

try:
    import pywt
    PYWT_AVAILABLE = True
except ImportError:
    PYWT_AVAILABLE = False


__all__ = ('PAD_MODES_ODL2PYWT', 'PYWT_SUPPORTED_MODES', 'PYWT_AVAILABLE',
           'pywt_wavelet', 'pywt_pad_mode', 'pywt_coeff_shapes',
           'pywt_flat_coeff_size',
           'pywt_flat_array_from_coeffs', 'pywt_coeffs_from_flat_array',
           'pywt_single_level_decomp', 'pywt_single_level_recon',
           'pywt_multi_level_decomp', 'pywt_multi_level_recon')


PAD_MODES_ODL2PYWT = {'constant': 'zero',
                      'periodic': 'periodic',
                      'symmetric': 'symmetric',
                      'order0': 'constant',
                      'order1': 'smooth',
                      'pywt_periodic': 'periodization',
                      # Upcoming version of Pywavelets adds this
                      # 'reflect': 'reflect'
                      }
PYWT_SUPPORTED_MODES = PAD_MODES_ODL2PYWT.values()


def pywt_wavelet(wavelet):
    """Convert ``wavelet`` to a `pywt.Wavelet` instance."""
    if isinstance(wavelet, pywt.Wavelet):
        return wavelet
    else:
        return pywt.Wavelet(wavelet)


def pywt_pad_mode(pad_mode, pad_const=0):
    """Convert ODL-style padding mode to pywt-style padding mode."""
    pad_mode = str(pad_mode).lower()
    if pad_mode == 'constant' and pad_const != 0.0:
        raise ValueError('constant padding with constant != 0 not supported '
                         'for `pywt` back-end')
    try:
        return PAD_MODES_ODL2PYWT[pad_mode]
    except KeyError:
        raise ValueError("`pad_mode` '{}' not understood".format(pad_mode))


def pywt_coeff_shapes(shape, wavelet, nlevels, mode):
    """Return a list of coefficient shapes in the specified transform.

    The wavelet transform specified by ``wavelet``, ``nlevels`` and
    ``mode`` produces a sequence of approximation and detail
    coefficients. This function computes the shape of those
    coefficient arrays and returns them as a list.

    Parameters
    ----------
    shape : sequence
        Shape of an input to the transform.
    wavelet : string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        If a string is given, it is converted to a `pywt.Wavelet`.
        Use `pywt.wavelist` to get a list of available wavelets.
    nlevels : positive int
        Number of scaling levels to be used in the decomposition. The
        maximum number of levels can be calculated with
        `pywt.dwt_max_level`.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.

    Returns
    -------
    shapes : list
        The shapes of the approximation and detail coefficients at
        different scaling levels in the following order:

            ``[shape_aN, shape_DN, ..., shape_D1]``

        Here, ``shape_aN`` is the shape of the N-th level approximation
        coefficient array, and ``shape_Di`` is the shape of all
        i-th level detail coefficients.

    See Also
    --------
    pywt.dwt_coeff_len : function used to determine coefficient sizes at
        each scaling level

    Examples
    --------
    Determine the coefficient shapes for 2 scaling levels in 3 dimensions
    with zero-padding. Approximation coefficient shape comes first, then
    the level-2 detail coefficient and last the level-1 detail coefficient:

    >>> pywt_coeff_shapes(shape=(16, 17, 18), wavelet='db2', nlevels=2,
    ...                   mode='zero')
    [(6, 6, 6), (6, 6, 6), (9, 10, 10)]

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html
    """
    shape = tuple(shape)
    if any(int(s) != s for s in shape):
        raise ValueError('`shape` may only contain integers, got {}'
                         ''.format(shape))
    wavelet = pywt_wavelet(wavelet)

    nlevels, nlevels_in = int(nlevels), nlevels
    if nlevels_in != nlevels:
        raise ValueError('`nlevels` must be integer, got {}'
                         ''.format(nlevels_in))
    # TODO: adapt for axes
    for i, n in enumerate(shape):
        max_levels = pywt.dwt_max_level(n, wavelet.dec_len)
        if max_levels == 0:
            raise ValueError('in axis {}: data size {} too small for '
                             'transform, results in maximal `nlevels` of 0'
                             ''.format(i, n))
        if not 0 < nlevels <= max_levels:
            raise ValueError('in axis {}: `nlevels` must satisfy 0 < nlevels '
                             '<= {}, got {}'
                             ''.format(i, max_levels, nlevels))

    mode, mode_in = str(mode).lower(), mode
    if mode not in PYWT_SUPPORTED_MODES:
        raise ValueError("mode '{}' not understood".format(mode_in))

    # Use pywt.dwt_coeff_len to determine the coefficient lengths at each
    # scale recursively. Start with the image shape and use the last created
    # shape for the next step.
    shape_list = [shape]
    for _ in range(nlevels):
        shape = tuple(pywt.dwt_coeff_len(n, filter_len=wavelet.dec_len,
                                         mode=mode)
                      for n in shape_list[-1])
        shape_list.append(shape)

    # Add a duplicate of the last entry for the approximation coefficients
    shape_list.append(shape_list[-1])

    # We created the list in reversed order, reverse it. Remove also the
    # superfluous image shape at the end.
    shape_list.reverse()
    shape_list.pop()
    return shape_list


def pywt_flat_coeff_size(shape, wavelet, nlevels, mode):
    """Return the size of a flat array containing all coefficients.

    The wavelet transform specified by ``wavelet``, ``nlevels`` and
    ``mode`` produces a sequence of approximation and detail
    coefficients. This function computes the total size of those
    coefficients when stored in a flat vector.

    Parameters
    ----------
    shape : sequence
        Shape of an input to the transform.
    wavelet : string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        If a string is given, it is converted to a `pywt.Wavelet`.
        Use `pywt.wavelist` to get a list of available wavelets.
    nlevels : positive int
        Number of scaling levels to be used in the decomposition. The
        maximum number of levels can be calculated with
        `pywt.dwt_max_level`.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.

    Returns
    -------
    flat_size : int
        Size of a flat array containing all approximation and detail
        coefficients for the specified transform.

    See Also
    --------
    pywt_coeff_shapes : calculate the shapes of coefficients in a
        specified wavelet transform

    Examples
    --------
    Determine the total size of a flat coefficient array for 2 scaling
    levels in 3 with zero-padding:

    >>> pywt_flat_coeff_size(shape=(16, 17, 18), wavelet='db2', nlevels=2,
    ...                      mode='zero')
    8028
    >>> 16 * 17 * 18  # original size, smaller -> redundancy in transform
    4896

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html
    """
    shapes = pywt_coeff_shapes(shape, wavelet, nlevels, mode)
    # 1 x approx coeff and (2**n - 1) x detail coeff
    ndim = len(shapes[0])
    return (np.prod(shapes[0]) +
            (2 ** ndim - 1) * sum(np.prod(shape)
                                  for shape in shapes[1:]))


def pywt_flat_array_from_coeffs(coeffs):
    """Return a flat array from a Pywavelets coefficient sequence.

    Parameters
    ----------
    coeffs : ordered sequence
        Sequence of approximation and detail coefficients in the format

            ``[aN, DN, ... D1]``,

        where ``aN`` is the N-th level approximation coefficient array and
        ``Di`` the tuple of i-th level detail coefficient arrays.

    Returns
    -------
    arr : `numpy.ndarray`
        Flat coefficient vector containing approximation and detail
        coefficients in the same order as they appear in ``coeffs``.

    See Also
    --------
    pywt_coeffs_from_flat_array : Conversion from flat array to
        coefficient list, the inverse of this function.

    Examples
    --------
    Flatten a list of 1 approximation and ``2 ** 2 - 1 = 3`` detail
    coefficients (sizes not representative):

    >>> approx = [[1, 1]]  # shape (1, 2)
    >>> details = ([[2, 2]], [[3, 3]], [[4, 4]])  # 3 * shape (1, 2)
    >>> pywt_flat_array_from_coeffs([approx, details])
    array([1, 1, 2, 2, 3, 3, 4, 4])
    """
    # We convert to numpy array since we need the sizes anyway, and np.size
    # is as expensive as np.asarray().size.
    approx = np.asarray(coeffs[0])
    details_list = []
    flat_sizes = [approx.size]

    for details in coeffs[1:]:
        if isinstance(details, tuple):
            detail_arrs = tuple(np.asarray(detail) for detail in details)
        else:
            # Single detail coefficient array
            detail_arrs = (np.asarray(details),)

        flat_sizes.append(detail_arrs[0].size)
        details_list.append(detail_arrs)

    ndim = approx.ndim
    dtype = approx.dtype
    dcoeffs_per_scale = 2 ** ndim - 1

    flat_total_size = flat_sizes[0] + dcoeffs_per_scale * sum(flat_sizes[1:])
    flat_coeff = np.empty(flat_total_size, dtype=dtype)

    start = 0
    stop = flat_sizes[0]
    flat_coeff[start:stop] = approx.ravel()

    for fsize, details in zip(flat_sizes[1:], details_list):
        for detail in details:
            start, stop = stop, stop + fsize
            flat_coeff[start:stop] = detail.ravel()

    return flat_coeff


def pywt_coeffs_from_flat_array(arr, shapes):
    """Convert a flat array into a ``pywt`` coefficient list.

    Parameters
    ----------
    arr : `array-like`
        A flat coefficient vector containing approximation and detail
        coefficients with order and sizes determined by ``shapes``.
    shapes : sequence
        The shapes of the approximation and detail coefficients at
        different scaling levels in the following order:

            ``[shape_aN, shape_DN, ..., shape_D1]``

        Here, ``shape_aN`` is the shape of the N-th level approximation
        coefficient array, and ``shape_Di`` is the shape of all
        i-th level detail coefficients.

    Returns
    -------
    coeff_list : structured list
        List of approximation and detail coefficients in the format

            ``[aN, DN, ... D1]``,

        where ``aN`` is the N-th level approximation coefficient array and
        ``Di`` the tuple of i-th level detail coefficient arrays. Each of
        the ``Di`` tuples has length ``2 ** ndim - 1``, where ``ndim`` is
        the number of dimensions of ``arr``.

    See Also
    --------
    pywt_flat_array_from_coeffs : Conversion from coefficient list to
        flat array, i.e. the inverse of this function.

    Examples
    --------
    Turn the flat array from the example in `pywt_flat_array_from_coeffs`
    back into a coefficient list:

    >>> arr = [1, 1, 2, 2, 3, 3, 4, 4]
    >>> # approximation and detail coefficient shapes
    >>> shapes = [(1, 2), (1, 2)]
    >>>
    >>> coeffs = pywt_coeffs_from_flat_array(arr, shapes)
    >>> approx, details = coeffs
    >>> print(approx)
    [[1 1]]
    >>> print(details)
    (array([[2, 2]]), array([[3, 3]]), array([[4, 4]]))
    """
    arr = np.asarray(arr)
    flat_sizes = [np.prod(shp) for shp in shapes]
    start = 0
    stop = flat_sizes[0]
    coeff_list = [arr[start:stop].reshape(shapes[0])]
    ndim = len(shapes[0])
    dcoeffs_per_scale = 2 ** ndim - 1

    for fsize, shape in zip(flat_sizes[1:], shapes[1:]):
        start, stop = stop, stop + dcoeffs_per_scale * fsize

        detail_coeffs = tuple(
            c.reshape(shape)
            for c in np.split(arr[start:stop], dcoeffs_per_scale))

        coeff_list.append(detail_coeffs)

    return coeff_list


def pywt_dict_keys(length):
    """Return a list of coefficient dictionary keys as used in ``pywt``.

    The dictionary keys are composed of the characters ``'a'`` and ``'d'``
    and stand for "approximation" and "detail" coefficients. The letter
    in the i-th position of a key string signalizes which filtering
    operation the input undergoes in the i-th coordinate axis. For example,
    ``'ada'`` stands for low-pass filtering (-> approximation) in the
    first and last axes and high-pass filtering (-> detail) in the
    middle axis.

    Examples
    --------
    There are 8 variants for ``length=3``:

    >>> pywt_dict_keys(length=3)
    ['aaa', 'aad', 'ada', 'add', 'daa', 'dad', 'dda', 'ddd']
    """
    return list(''.join(k) for k in product('ad', repeat=length))


def pywt_single_level_decomp(arr, wavelet, mode):
    """Return single level wavelet decomposition coefficients from ``arr``.

    Parameters
    ----------
    arr : `array-like`
        Input array to the wavelet decomposition.
    wavelet : string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        If a string is given, it is converted to a `pywt.Wavelet`.
        Use `pywt.wavelist` to get a list of available wavelets.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.

    Returns
    -------
    approx : `numpy.ndarray`
        Approximation coefficients, a single array.
    details : tuple of `numpy.ndarray`'s
        Detail coefficients, ``2 ** ndim - 1`` arrays, where ``ndim``
        is the number of dimensions of ``arr``.

    See Also
    --------
    pywt_single_level_recon : Single-level reconstruction, i.e. the
        inverse of this function.
    pywt_multi_level_decomp : Multi-level version of the decompostion.

    Examples
    --------
    Decomposition of a small two-dimensional array using a Haar wavelet
    and zero padding:

    >>> arr = [[1, 1, 1],
    ...        [1, 0, 0],
    ...        [0, 1, 1]]
    >>> coeffs = pywt_single_level_decomp(arr, wavelet='haar', mode='zero')
    >>> approx, details = coeffs
    >>> print(approx)
    [[ 1.5  0.5]
     [ 0.5  0.5]]
    >>> for detail in details:
    ...     print(detail)
    [[ 0.5  0.5]
     [-0.5  0.5]]
    [[ 0.5  0.5]
     [ 0.5  0.5]]
    [[-0.5  0.5]
     [-0.5  0.5]]

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html
    """
    # Handle input
    arr = np.asarray(arr)
    wavelet = pywt_wavelet(wavelet)
    mode, mode_in = str(mode).lower(), mode
    if mode not in PYWT_SUPPORTED_MODES:
        raise ValueError("mode '{}' not understood".format(mode_in))

    # Compute single level DWT using pywt.dwtn and pick the approximation
    # and detail coefficients from the dictionary
    coeff_dict = pywt.dwtn(arr, wavelet, mode)
    dict_keys = pywt_dict_keys(arr.ndim)
    approx = coeff_dict[dict_keys[0]]
    details = tuple(coeff_dict[k] for k in dict_keys[1:])
    return approx, details


def pywt_single_level_recon(approx, details, wavelet, mode, recon_shape=None):
    """Return single level wavelet reconstruction from given coefficients.

    Parameters
    ----------
    approx : `array-like`
        Approximation coefficients.
    details : sequence of `array-like`'s
        Detail coefficients. The length of the sequence must be
        ``2 ** ndim - 1``, where ``ndim`` is the number of dimensions
        in ``approx``.
    wavelet :  string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        Use `pywt.wavelist` to get a list of available wavelets.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.
    recon_shape : sequence of ints, optional
        Shape of the array to be reconstructed. Without this parameter,
        the reconstructed array always has even shape due to upsampling
        by a factor of 2. To get reconstructions with odd shapes, this
        parameter is required.

    Returns
    -------
    recon : `numpy.ndarray`
        The single-level wavelet reconstruction.

    See Also
    --------
    pywt_single_level_decomp : Single-level decomposition, i.e. the
        inverse of this function.
    pywt_multi_level_recon : Multi-level version of the reconstruction.

    Examples
    --------
    Take the coefficients from the example in `pywt_single_level_decomp`
    and reconstruct the original array. Without ``recon_shape``, we get
    a ``(4, 4)`` array as reconstruction:

    >>> approx = [[1.5, 0.5],
    ...           [0.5, 0.5]]
    >>>
    >>> details = ([[ 0.5, 0.5],
    ...             [-0.5, 0.5]],
    ...            [[ 0.5, 0.5],
    ...             [ 0.5, 0.5]],
    ...            [[-0.5, 0.5],
    ...             [-0.5, 0.5]])
    >>> # Gives even shape by default
    >>> pywt_single_level_recon(approx, details, wavelet='haar', mode='zero')
    array([[ 1.,  1.,  1.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 0.,  1.,  1.,  0.],
           [ 0.,  0.,  0.,  0.]])
    >>> # Original shape can only be recovered if given explicitly
    >>> # in this case
    >>> pywt_single_level_recon(approx, details, wavelet='haar', mode='zero',
    ...                         recon_shape=(3, 3))
    array([[ 1.,  1.,  1.],
           [ 1.,  0.,  0.],
           [ 0.,  1.,  1.]])

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html
    """
    # Handle input
    approx = np.asarray(approx)
    if len(details) != 2 ** approx.ndim - 1:
        raise ValueError('`details` must be a sequence of length {}, got '
                         'length {}'
                         .format(2 ** approx.ndim - 1, len(details)))
    details = tuple(np.asarray(detail) for detail in details)

    if recon_shape is not None:
        recon_shape, recon_shape_in = tuple(recon_shape), recon_shape
        if any(int(s) != s for s in recon_shape):
            raise ValueError('`recon_shape` may only contain integers, got {}'
                             ''.format(recon_shape_in))

    wavelet = pywt_wavelet(wavelet)
    mode, mode_in = str(mode).lower(), mode
    if mode not in PYWT_SUPPORTED_MODES:
        raise ValueError("mode '{}' not understood".format(mode_in))

    coeff_dict = {}
    dict_keys = pywt_dict_keys(np.ndim(approx))
    if len(details) != len(dict_keys) - 1:
        raise ValueError('wrong number of detail coefficients: expected {}, '
                         'got {}'.format(len(dict_keys) - 1, len(details)))
    coeff_dict[dict_keys[0]] = approx
    coeff_dict.update(zip(dict_keys[1:], details))
    recon = pywt.idwtn(coeff_dict, wavelet, mode)

    if recon_shape is not None:
        recon_slc = []
        for i, (n_recon, n_intended) in enumerate(zip(recon.shape,
                                                      recon_shape)):
            if n_recon == n_intended + 1:
                # Upsampling added one entry too much in this axis, drop
                # last one
                recon_slc.append(slice(-1))
            elif n_recon == n_intended:
                recon_slc.append(slice(None))
            else:
                raise ValueError('in axis {}: expected size {} or {} in '
                                 '`recon_shape`, got {}'
                                 ''.format(i, n_recon - 1, n_recon,
                                           n_intended))

        recon = recon[tuple(recon_slc)]

    return recon


def pywt_multi_level_decomp(arr, wavelet, nlevels, mode):
    """Return multi-level wavelet decomposition coefficients from ``arr``.

    Parameters
    ----------
    arr : `array-like`
        Input array to the wavelet decomposition.
    wavelet :  string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        Use `pywt.wavelist` to get a list of available wavelets.
    nlevels : positive int
        Number of scaling levels to be used in the decomposition. The
        maximum number of levels can be calculated with
        `pywt.dwt_max_level`.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.

    Returns
    -------
    coeff_list : structured list
        List of approximation and detail coefficients in the format

            ``[aN, DN, ... D1]``,

        where ``aN`` is the N-th level approximation coefficient array and
        ``Di`` the tuple of i-th level detail coefficient arrays. Each of
        the ``Di`` tuples has length ``2 ** ndim - 1``, where ``ndim`` is
        the number of dimensions of the input array.
        See the documentation for the `multilevel decomposition`_ in
        PyWavelets for more information.

    See Also
    --------
    pywt_single_level_decomp : Single-level version of this function.
    pywt_multi_level_recon : Multi-level reconstruction, i.e. the inverse
        of this function.

    Examples
    --------
    Decomposition of a ``(4, 4)`` array using Haar wavelet, zero-padding
    and 2 scaling levels:

    >>> arr = [[1, 1, 0, 0],
    ...        [0, 0, 0, 1],
    ...        [1, 1, 1, 1],
    ...        [0, 1, 1, 0]]
    >>> coeffs = pywt_multi_level_decomp(arr, 'haar', 2, 'zero')
    >>> # "2" -> scaling level 2, "1" -> scaling level 1
    >>> approx2, details2, details1 = coeffs
    >>> print(approx2)
    [[ 2.25]]
    >>> for detail in details2:
    ...     print(detail)
    [[ 0.25]]
    [[-0.75]]
    [[ 0.25]]
    >>> for detail in details1:
    ...     print(detail)
    [[ 0.  -0.5]
     [-0.5  0.5]]
    [[ 1.  -0.5]
     [ 0.5  0.5]]
    [[ 0.   0.5]
     [ 0.5 -0.5]]

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html

    .. _multilevel decomposition:
       https://pywavelets.readthedocs.io/en/latest/ref/dwt-discrete-\
wavelet-transform.html#multilevel-decomposition-using-wavedec
    """
    # Handle input
    arr = np.asarray(arr)
    wavelet = pywt_wavelet(wavelet)

    # TODO: adapt for axes
    nlevels, nlevels_in = int(nlevels), nlevels
    if nlevels_in != nlevels:
        raise ValueError('`nlevels` must be integer, got {}'
                         ''.format(nlevels_in))
    for i, n in enumerate(arr.shape):
        max_levels = pywt.dwt_max_level(n, wavelet.dec_len)
        if max_levels == 0:
            raise ValueError('in axis {}: data size {} too small for '
                             'transform, results in maximal `nlevels` of 0'
                             ''.format(i, n))
        if not 0 < nlevels <= max_levels:
            raise ValueError('in axis {}: `nlevels` must satisfy 0 < nlevels '
                             '<= {}, got {}'
                             ''.format(i, max_levels, nlevels))

    mode, mode_in = str(mode).lower(), mode
    if mode not in PYWT_SUPPORTED_MODES:
        raise ValueError("mode '{}' not understood".format(mode_in))

    # Fill the list with detail coefficients from coarsest to finest level,
    # by recursively applying the single-level transform to the approximation
    # coefficients, starting with the input array. Append the final
    # approximation coefficients.
    coeff_list = []
    approx, details = pywt_single_level_decomp(arr, wavelet, mode)
    coeff_list.append(details)

    for _ in range(1, nlevels):
        approx, details = pywt_single_level_decomp(approx, wavelet, mode)
        coeff_list.append(details)

    coeff_list.append(approx)
    coeff_list.reverse()

    return coeff_list


def pywt_multi_level_recon(coeff_list, wavelet, mode, recon_shape=None):
    """Return multi-level wavelet decomposition coefficients from ``arr``.

    Parameters
    ----------
    coeff_list : structured list
        List of approximation and detail coefficients in the format

            ``[aN, DN, ... D1]``,

        where ``aN`` is the N-th level approximation coefficient array and
        ``Di`` the tuple of i-th level detail coefficient arrays. Each of
        the ``Di`` tuples must have length ``2 ** ndim - 1``, where ``ndim``
        is the number of dimensions of the input array.
    wavelet :  string or `pywt.Wavelet`
        Specification of the wavelet to be used in the transform.
        Use `pywt.wavelist` to get a list of available wavelets.
    mode : string, optional
        PyWavelets style signal extension mode. See `signal extension modes`_
        for available options.
    recon_shape : sequence of ints, optional
        Shape of the array to be reconstructed. Without this parameter,
        the reconstructed array always has even shape due to upsampling
        by a factor of 2. To get reconstructions with odd shapes, this
        parameter is required.

    Returns
    -------
    recon : `numpy.ndarray`
        Wavelet reconstruction from the given coefficients.

    See Also
    --------
    pywt_single_level_recon : Single-level version of this function.
    pywt_multi_level_decomp : Multi-level decomposition, i.e. the inverse
        of this function.

    Examples
    --------
    Reconstruct the original array from the decomposition in the example
    in `pywt_multi_level_decomp`:

    >>> orig_arr = [[1, 1, 0, 0],
    ...             [0, 0, 0, 1],
    ...             [1, 1, 1, 1],
    ...             [0, 1, 1, 0]]
    >>> approx2 = [[2.25]]
    >>> details2 = ([[0.25]], [[-0.75]], [[0.25]])
    >>> details1 = ([[0, -0.5],
    ...              [-0.5, 0.5]],
    ...             [[1, -0.5],
    ...              [0.5, 0.5]],
    ...             [[0, 0.5],
    ...              [0.5, -0.5]])
    >>> coeffs = [approx2, details2, details1]
    >>> recon = pywt_multi_level_recon(coeffs, wavelet='haar', mode='zero')
    >>> np.allclose(recon, orig_arr)
    True

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-\
modes.html
    """
    if recon_shape is not None:
        recon_shape, recon_shape_in = tuple(recon_shape), recon_shape
        if any(int(s) != s for s in recon_shape):
            raise ValueError('`recon_shape` may only contain integers, got {}'
                             ''.format(recon_shape_in))

    recon = np.asarray(coeff_list[0])

    for cur_details, next_details in zip(coeff_list[1:-1], coeff_list[2:]):
        if isinstance(next_details, tuple):
            next_shape = np.shape(next_details[0])
        else:
            # Single details array
            next_shape = np.shape(next_details)

        recon = pywt_single_level_recon(recon, cur_details, wavelet, mode,
                                        recon_shape=next_shape)

    # Last reco step uses `recon_shape` for shape correction
    return pywt_single_level_recon(recon, coeff_list[-1], wavelet, mode,
                                   recon_shape=recon_shape)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests(skip_if=not PYWT_AVAILABLE)



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Phantoms given by simple geometric objects such as cubes or spheres."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np

__all__ = ('cuboid', 'ellipse_phantom', 'indicate_proj_axis')


def cuboid(space, min_pt=None, max_pt=None):
    """Rectangular cuboid.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Discretized space in which the phantom is supposed to be created.
    min_pt : array-like of shape ``(space.ndim,)``, optional
        Lower left corner of the cuboid. If ``None`` is given, a quarter
        of the extent from ``space.min_pt`` towards the inside is chosen.
    min_pt : array-like of shape ``(space.ndim,)``, optional
        Upper right corner of the cuboid. If ``None`` is given, ``min_pt``
        plus half the extent is chosen.

    Returns
    -------
    phantom : `DiscretizedSpaceElement`
        The generated cuboid phantom in ``space``.

    Examples
    --------
    If both ``min_pt`` and ``max_pt`` are omitted, the cuboid lies in the
    middle of the space domain and extends halfway towards all sides:

    >>> space = odl.uniform_discr([0, 0], [1, 1], [4, 6])
    >>> print(odl.phantom.cuboid(space))
    [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
     [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
     [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
     [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

    By specifying the corners, the cuboid can be arbitrarily shaped:

    >>> print(odl.phantom.cuboid(space, [0.25, 0], [0.75, 0.5]))
    [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
     [1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
     [1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
     [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]
    """
    dom_min_pt = np.asarray(space.domain.min())
    dom_max_pt = np.asarray(space.domain.max())

    if min_pt is None:
        min_pt = dom_min_pt * 0.75 + dom_max_pt * 0.25
    if max_pt is None:
        max_pt = dom_min_pt * 0.25 + dom_max_pt * 0.75

    min_pt = np.atleast_1d(min_pt)
    max_pt = np.atleast_1d(max_pt)

    if min_pt.shape != (space.ndim,):
        raise ValueError('shape of `min_pt` must be {}, got {}'
                         ''.format((space.ndim,), min_pt.shape))
    if max_pt.shape != (space.ndim,):
        raise ValueError('shape of `max_pt` must be {}, got {}'
                         ''.format((space.ndim,), max_pt.shape))

    def phantom(x):
        result = True

        for xi, xmin, xmax in zip(x, min_pt, max_pt):
            result = (result &
                      np.less_equal(xmin, xi) & np.less_equal(xi, xmax))
        return result

    return space.element(phantom)


def indicate_proj_axis(space, scale_structures=0.5):
    """Phantom indicating along which axis it is projected.

    The number (n) of rectangles in a parallel-beam projection along a main
    axis (0, 1, or 2) indicates the projection to be along the (n-1)the
    dimension.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Discretized space in which the phantom is supposed to be created
    scale_structures : positive float in (0, 1]
        Scales objects (cube, cuboids)

    Returns
    -------
    phantom : ``space`` element
        Projection helper phantom in ``space``.

    Examples
    --------
    >>> space = odl.uniform_discr([0] * 3, [1] * 3, [8, 8, 8])
    >>> phan = indicate_proj_axis(space).asarray()
    >>> print(np.sum(phan, 0))
    [[ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  3.  3.  0.  0.  0.]
     [ 0.  0.  0.  3.  3.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]]
    >>> print(np.sum(phan, 1))
    [[ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  2.  2.  0.  0.  0.]
     [ 0.  0.  0.  2.  2.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  1.  1.  0.  0.  0.]
     [ 0.  0.  0.  1.  1.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]]
    >>> print(np.sum(phan, 2))
    [[ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  2.  2.  0.  0.  0.]
     [ 0.  0.  0.  2.  2.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  2.  0.  0.  0.]
     [ 0.  0.  0.  2.  0.  0.  0.  0.]
     [ 0.  0.  0.  0.  0.  0.  0.  0.]]
    """
    if not 0 < scale_structures <= 1:
        raise ValueError('`scale_structures` ({}) is not in (0, 1]'
                         ''.format(scale_structures))

    assert space.ndim == 3

    shape = space.shape
    phan = np.zeros(shape)
    shape = np.array(shape) - 1
    cen = np.round(0.5 * shape)
    dx = np.floor(scale_structures * 0.25 * shape)
    dx[dx == 0] = 1

    # cube of size 2 * dx
    x0 = (cen - 3 * dx)[0]
    x, y, z = cen - 1 * dx
    phan[x0:x, y:-y, z:-z] = 1

    # 1st cuboid of size (dx[0], dx[1], 2 * dx[2])
    x0 = (cen + 1 * dx)[1]
    x1 = (cen + 2 * dx)[1]
    y0 = cen[1]
    z = (cen - dx)[2]
    phan[x0:x1, y0:-y, z:-z] = 1

    # 2nd cuboid of (dx[0], dx[1], 2 * dx[2]) touching the first diagonally
    # at a long edge
    x0 = (cen + 2 * dx)[1]
    x1 = (cen + 3 * dx)[1]
    y1 = cen[1]
    z = (cen - dx)[2]
    phan[x0:x1, y:y1, z:-z] = 1

    return space.element(phan)


def _getshapes_2d(center, max_radius, shape):
    """Calculate indices and slices for the bounding box of a disk."""
    index_mean = shape * center
    index_radius = max_radius / 2.0 * np.array(shape)

    min_idx = np.floor(index_mean - index_radius).astype(int)
    max_idx = np.ceil(index_mean + index_radius).astype(int)
    idx = [slice(minx, maxx) for minx, maxx in zip(min_idx, max_idx)]
    shapes = [(idx[0], slice(None)),
              (slice(None), idx[1])]
    return idx, shapes


def _ellipse_phantom_2d(space, ellipses):
    """Create a phantom of ellipses in 2d space.

    Parameters
    ----------
    space : `DiscreteLp`
        Space the phantom should be generated in.
    ellipses : list of lists
        Each row should contain:
        'value', 'axis_1', 'axis_2', 'center_x', 'center_y', 'rotation'
        The ellipses should be contained the he rectangle [-1, -1] x [1, 1].

    Returns
    -------
    phantom : ``space`` element
        2D ellipse phantom in ``space``.

    See Also
    --------
    shepp_logan : The typical use-case for this function.
    """

    # Blank image
    p = np.zeros(space.shape, dtype=space.dtype)

    # Create the pixel grid
    grid_in = space.grid.meshgrid
    minp = space.grid.min()
    maxp = space.grid.max()

    # move points to [-1, 1]
    grid = []
    for i in range(2):
        meani = (minp[i] + maxp[i]) / 2.0
        diffi = (maxp[i] - minp[i]) / 2.0
        grid += [(grid_in[i] - meani) / diffi]

    for ellip in ellipses:
        assert len(ellip) == 6

        intensity = ellip[0]
        a_squared = ellip[1] ** 2
        b_squared = ellip[2] ** 2
        x0 = ellip[3]
        y0 = ellip[4]
        theta = ellip[5] * np.pi / 180

        scales = [1 / a_squared, 1 / b_squared]
        center = (np.array([x0, y0]) + 1.0) / 2.0

        # Create the offset x,y and z values for the grid
        if theta != 0:
            # Rotate the points to the expected coordinate system.
            ctheta = np.cos(theta)
            stheta = np.sin(theta)

            mat = np.array([[ctheta, stheta],
                            [-stheta, ctheta]])

            # Calculate the points that could possibly be inside the volume
            # Since the points are rotated, we cannot do anything directional
            # without more logic
            max_radius = np.sqrt(
                np.abs(mat).dot([a_squared, b_squared]))
            idx, shapes = _getshapes_2d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            offset_points = [vec * (xi - x0i)[..., np.newaxis]
                             for xi, vec, x0i in zip(subgrid,
                                                     mat.T,
                                                     [x0, y0])]
            rotated = offset_points[0] + offset_points[1]
            np.square(rotated, out=rotated)
            radius = np.dot(rotated, scales)
        else:
            # Calculate the points that could possibly be inside the volume
            max_radius = np.sqrt([a_squared, b_squared])
            idx, shapes = _getshapes_2d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            squared_dist = [ai * (xi - x0i) ** 2
                            for xi, ai, x0i in zip(subgrid,
                                                   scales,
                                                   [x0, y0])]

            # Parentheses to get best order for broadcasting
            radius = squared_dist[0] + squared_dist[1]

        # Find the pixels within the ellipse
        inside = radius <= 1

        # Add the ellipse intensity to those pixels
        p[idx][inside] += intensity

    return space.element(p)


def _getshapes_3d(center, max_radius, shape):
    """Calculate indices and slices for the bounding box of a ball."""
    index_mean = shape * center
    index_radius = max_radius / 2.0 * np.array(shape)

    min_idx = np.floor(index_mean - index_radius).astype(int)
    max_idx = np.ceil(index_mean + index_radius).astype(int)
    idx = [slice(minx, maxx) for minx, maxx in zip(min_idx, max_idx)]
    shapes = [(idx[0], slice(None), slice(None)),
              (slice(None), idx[1], slice(None)),
              (slice(None), slice(None), idx[2])]
    return idx, shapes


def _ellipse_phantom_3d(space, ellipses):
    """Create an ellipse phantom in 3d space.

    Parameters
    ----------
    space : `DiscreteLp`
        Space the phantom should be generated in.
    ellipses : list of lists
        Each row should contain:
        'value', 'axis_1', 'axis_2', 'axis_3',
        'center_x', 'center_y', 'center_z',
        'rotation_phi', 'rotation_theta', 'rotation_psi'
        The ellipses should be contained the he rectangle
        [-1, -1, -1] x [1, 1, 1].

    Returns
    -------
    phantom : ``space`` element
        3D ellipse phantom in ``space``.

    See Also
    --------
    shepp_logan : The typical use-case for this function.
    """

    # Blank image
    p = np.zeros(space.shape)

    # Create the pixel grid
    grid_in = space.grid.meshgrid
    minp = space.grid.min()
    maxp = space.grid.max()

    # move points to [-1, 1]
    grid = []
    for i in range(3):
        meani = (minp[i] + maxp[i]) / 2.0
        diffi = (maxp[i] - minp[i]) / 2.0
        grid += [(grid_in[i] - meani) / diffi]

    for ellip in ellipses:
        assert len(ellip) == 10

        intensity = ellip[0]
        a_squared = ellip[1] ** 2
        b_squared = ellip[2] ** 2
        c_squared = ellip[3] ** 2
        x0 = ellip[4]
        y0 = ellip[5]
        z0 = ellip[6]
        phi = ellip[7] * np.pi / 180
        theta = ellip[8] * np.pi / 180
        psi = ellip[9] * np.pi / 180

        scales = [1 / a_squared, 1 / b_squared, 1 / c_squared]
        center = (np.array([x0, y0, z0]) + 1.0) / 2.0

        # Create the offset x,y and z values for the grid
        if any([phi, theta, psi]):
            # Rotate the points to the expected coordinate system.
            cphi = np.cos(phi)
            sphi = np.sin(phi)
            ctheta = np.cos(theta)
            stheta = np.sin(theta)
            cpsi = np.cos(psi)
            spsi = np.sin(psi)

            mat = np.array([[cpsi * cphi - ctheta * sphi * spsi,
                             cpsi * sphi + ctheta * cphi * spsi,
                             spsi * stheta],
                            [-spsi * cphi - ctheta * sphi * cpsi,
                             -spsi * sphi + ctheta * cphi * cpsi,
                             cpsi * stheta],
                            [stheta * sphi,
                             -stheta * cphi,
                             ctheta]])

            # Calculate the points that could possibly be inside the volume
            # Since the points are rotated, we cannot do anything directional
            # without more logic

            max_radius = np.sqrt(
                np.abs(mat).dot([a_squared, b_squared, c_squared]))
            idx, shapes = _getshapes_3d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            offset_points = [vec * (xi - x0i)[..., np.newaxis]
                             for xi, vec, x0i in zip(subgrid,
                                                     mat.T,
                                                     [x0, y0, z0])]
            rotated = offset_points[0] + offset_points[1] + offset_points[2]
            np.square(rotated, out=rotated)
            radius = np.dot(rotated, scales)
        else:
            # Calculate the points that could possibly be inside the volume
            max_radius = np.sqrt([a_squared, b_squared, c_squared])
            idx, shapes = _getshapes_3d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            squared_dist = [ai * (xi - x0i) ** 2
                            for xi, ai, x0i in zip(subgrid,
                                                   scales,
                                                   [x0, y0, z0])]

            # Parentheses to get best order for broadcasting
            radius = squared_dist[0] + (squared_dist[1] + squared_dist[2])

        # Find the pixels within the ellipse
        inside = radius <= 1

        # Add the ellipse intensity to those pixels
        p[idx][inside] += intensity

    return space.element(p)


def ellipse_phantom(space, ellipses):
    """Return a phantom given by ellipses.

    Parameters
    ----------
    space : `DiscreteLp`
        Space in which the phantom is created, must be 2- or 3-dimensional.
    ellipses : sequence of sequences
        If ``space`` is 2-dimensional each row should contain:

        'value', 'axis_1', 'axis_2', 'center_x', 'center_y', 'rotation'

        If ``space`` is 3-dimensional each row should contain:

        'value', 'axis_1', 'axis_2', 'axis_3',
        'center_x', 'center_y', 'center_z',
        'rotation_phi', 'rotation_theta', 'rotation_psi'

        The ellipses need to be given such that the ellipses fall in the
        rectangle [-1, -1] x [1, 1] or equivalent in 3d.

    Notes
    -----
    The phantom is created by adding the values of each ellipse. The ellipses
    are defined by a center point (center_x, center_y, [center_z]) and the
    length of its principial axes (axis_1, axis_2, [axis_2]) and euler angles.

    This function is heavily optimized, achieving runtimes about 20 times
    faster than "trivial" implementations. It is therefore recommended to use
    it in all phantoms where applicable.

    The main optimization is that it only considers a subset of all the
    points when updating for each ellipse. It does this by first finding
    a subset of points that could possibly be inside the ellipse. This
    optimization is very good for "spherical" ellipsoids, but not so
    much for elongated or rotated ones.

    It also does calculations wherever possible on the meshgrid instead of
    individual points.

    Examples
    --------
    Create a circle with a smaller circle inside:

    >>> space = odl.uniform_discr([-1, -1], [1, 1], [5, 5])
    >>> ellipses = [[1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
    ...             [1.0, 0.6, 0.6, 0.0, 0.0, 0.0]]
    >>> print(ellipse_phantom(space, ellipses))
    [[0.0, 0.0, 1.0, 0.0, 0.0],
     [0.0, 1.0, 2.0, 1.0, 0.0],
     [1.0, 2.0, 2.0, 2.0, 1.0],
     [0.0, 1.0, 2.0, 1.0, 0.0],
     [0.0, 0.0, 1.0, 0.0, 0.0]]

    See Also
    --------
    odl.phantom.transmission.shepp_logan : Classical Shepp-Logan phantom,
        typically used for transmission imaging
    odl.phantom.transmission.shepp_logan_ellipses : Ellipses for the
        Shepp-Logan phantom
    """

    if space.ndim == 2:
        return _ellipse_phantom_2d(space, ellipses)
    elif space.ndim == 3:
        return _ellipse_phantom_3d(space, ellipses)
    else:
        raise ValueError('dimension not 2 or 3, no phantom available')


if __name__ == '__main__':
    # Show the phantoms
    import odl

    # cuboid 1D
    discr = odl.uniform_discr(-1, 1, 300)
    cuboid(discr).show('cuboid 1d')

    # cuboid 2D
    discr = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    cuboid(discr).show('cuboid 2d')

    # cuboid 3D
    discr = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    cuboid(discr).show('cuboid 3d')
    indicate_proj_axis(discr).show('indicate_proj_axis 3d')

    # ellipse phantom 2D
    discr = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    ellipses = [[1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.6, 0.6, 0.0, 0.0, 0.0]]
    ellipse_phantom(discr, ellipses).show('ellipse phantom 2d')

    # ellipse phantom 3D
    discr = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    ellipses = [[1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]
    ellipse_phantom(discr, ellipses).show('ellipse phantom 3d')

    # Run also the doctests
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()
